########################################################################
# $Header: /var/local/cvsroot/4Suite/Ft/Lib/Resolvers.py,v 1.3 2005-01-05 10:04:22 mbrown Exp $
"""
Specialized and useful input source and URI tools

Copyright 2008-2015 Uche Ogbuji
"""

import os, sys
import io
import urllib, urllib.request
import email
from email.utils import formatdate as _formatdate

import amara3
IriError = amara3.iri.IriError
from amara3.iri import *

__all__ = [
'DEFAULT_URI_SCHEMES',
'DEFAULT_RESOLVER',
'scheme_registry_resolver',
'facade_resolver',
'uridict',
'resolver',
]

# URI schemes supported by resolver_base
DEFAULT_URI_SCHEMES = ('http', 'file', 'ftp', 'data', 'pkgdata')
if hasattr(urllib.request, 'HTTPSHandler'):
    DEFAULT_URI_SCHEMES += ('https',)
DEFAULT_HIERARCHICAL_SEP = '/' #a separator to place between path segments when creating URLs

class resolver:
    """
    """
    _supported_schemes = DEFAULT_URI_SCHEMES
    def __init__(self, authorizations=None, lenient=True):
        """
        """
        self.authorizations = authorizations
        self.lenient = lenient

    def resolve(self, uriRef, baseUri=None):
        """
        Takes a URI or a URI reference plus a base URI, produces a absolutized URI
        if a base URI was given, then attempts to obtain access to an entity
        representing the resource identified by the resulting URI,
        returning the entity as a stream (a file-like object).

        Raises a IriError if the URI scheme is unsupported or if a stream
        could not be obtained for any reason.
        """
        if not isinstance(uriRef, urllib.request.Request):
            if baseUri is not None:
                uri = self.absolutize(uriRef, baseUri)
                scheme = get_scheme(uri)
            else:
                uri = uriRef
                scheme = get_scheme(uriRef)
                # since we didn't use absolutize(), we need to verify here
                if scheme not in self._supported_schemes:
                    if scheme is None:
                        raise ValueError('When the URI to resolve is a relative '
                            'reference, it must be accompanied by a base URI.')
                    else:
                        raise IriError(IriError.UNSUPPORTED_SCHEME,
                                           scheme=scheme, resolver=self.__class__.__name__)
            req = urllib.request.Request(uri)
        else:
            req, uri = uriRef, uriRef.get_full_url()

        if self.authorizations and not self.authorize(uri):
            raise IriError(IriError.DENIED_BY_RULE, uri=uri)
        # Bypass urllib for opening local files.
        if scheme == 'file':
            path = uri_to_os_path(uri, attemptAbsolute=False)
            try:
                stream = open(path, 'rb')
            except IOError as e:
                raise IriError(IriError.RESOURCE_ERROR,
                                   loc='%s (%s)' % (uri, path),
                                   uri=uri, msg=str(e))
            # Add the extra metadata that urllib normally provides (sans
            # the poorly guessed Content-Type header).
            stats = os.stat(path)
            size = stats.st_size
            mtime = _formatdate(stats.st_mtime)
            headers = email.Message(io.StringIO(
                'Content-Length: %s\nLast-Modified: %s\n' % (size, mtime)))
            stream = urllib.addinfourl(stream, headers, uri)
        else:
            # urllib.request.urlopen, wrapped by us, will suffice for http, ftp,
            # data and gopher
            try:
                stream = urllib.request.urlopen(req)
            except IOError as e:
                raise IriError(IriError.RESOURCE_ERROR,
                                   uri=uri, loc=uri, msg=str(e))
        return stream

    def authorize(self, uri):
        """
        Implement an authorization mechanism for resolvers, allowing you to create "jails" where only certain URIs are allowed
        """
        for match, allow in self.authorizations:
            if callable(match):
                if match(uri):
                    return allow
            elif match:
                return allow
        #If authoriztions are specified, and none allow the URI, deny by default
        #The user can easily reverse this by adding an auth (True, True)
        return False

    def absolutize(self, uriRef, baseUri):
        """
        For most cases iri.absolutize is good enough, and does the main work of this function.
        
        Resolves a URI reference to absolute form, effecting the result of RFC
        3986 section 5. The URI reference is considered to be relative to
        the given base URI.

        Also verifies that the resulting URI reference has a scheme that
        resolve() supports, raising a IriError if it doesn't.

        Default implementation does not perform any validation on the base
        URI beyond that performed by iri.absolutize().

        If leniency has been turned on (self.lenient=True), accepts a base URI
        beginning with '/', in which case the argument is assumed to be an absolute
        path component of 'file' URI with no authority component.
        """
        # since we know how absolutize works, we can anticipate the scheme of
        # its return value and verify that it's supported first
        if self.lenient:
            # assume file: if leading "/"
            if baseUri.startswith('/'):
                baseUri = 'file://' + baseUri
        return absolutize(uriRef, baseUri, limit_schemes=self._supported_schemes)

DEFAULT_RESOLVER = resolver()


class scheme_registry_resolver(resolver):
    """
    Resolver that handles URI resolution with a registry for handling different
    URI schemes.  The default action if there is nothing registered for the scheme
    will be to fall back to base behavior *unless* you have in the mapping a special
    scheme None.  The callable object that is the value on that key will then be used
    as the default for all unknown schemes.

    The expected function signature for scheme call-backs matches
    inputsource.resolve, without the instance argument:

    resolve(uri, base=None)

    Reminder: Since this does not include self, if you are registering
    a method, use the method instance (i.e. myresolver().handler
    rather than myresolver.handler)

    You can manipulate the mapping directly using the "handlers" attribute.
    handlers - a Python dictionary with scheme names as keys (e.g. "http")
    and callable objects as values
    """
    def __init__(self, authorizations=None, lenient=True, handlers=None):
        """
        """
        self.lenient = lenient
        self.handlers = handlers or {}
        resolver.__init__(self, authorizations, lenient)

    def resolve(self, uri, base=None):
        scheme = get_scheme(uri)
        if not scheme:
            if base:
                scheme = get_scheme(base)
            if not scheme:
                #Another option is to fall back to Base class behavior
                raise Iri.IriError(Iri.IriError.SCHEME_REQUIRED,
                                   base=base, ref=uri)
        func = self.handlers.get(scheme)
        #import sys; print >> sys.stderr, (self, self.handlers)
        if not func:
            func = self.handlers.get(None)
            if not func:
                return resolver.resolve(self, uri, base)
        return func(uri, base)


#Eliminate this class by demonstrating a standard memoize tool with the resolver
class facade_resolver(resolver):
    """
    A type of resolver that uses a cache of resources, a dictionary of URI
    to result mappings (similar to memoizing the resolve method).  When a
    URI is provided for resolution, the mapping is first checked, and a
    stream is constructed by wrapping the mapping value string.
    If no match is found in the mapping, fall back to the base
    resolver logic.

    You can manipulate the mapping directly using the "cache" attribute.
    """
    def __init__(self, cache=None, observer=None):
        """
        cache - a dictionary with mapings from URI to value (as an object
        to be converted to a UTF-8 encoded string)
        observer - callable object invoked on each resolution request
        """
        default_resolver.__init__(self)
        self.cache = cache or {}
        self.observer = observer
        return

    def resolve(self, uri, base=None):
        self.observer(uri, base)
        #Does not factor in base.  Should it noramlize before checking?
        if uri in self.cache:
            cachedval = self.cache[uri]
            if isinstance(cachedval, unicode):
                return io.StringIO(cachedval.encode('utf-8'))
            else:
                return io.StringIO(str(cachedval))
        return default_resolver.resolve(self, uri, base)

#
class uridict(dict):
    """
    A dictionary that uses URIs as keys. It attempts to observe some degree of
    URI equivalence as defined in RFC 3986 section 6. For example, if URIs
    A and B are equivalent, a dictionary operation involving key B will return
    the same result as one involving key A, and vice-versa.

    This is useful in situations where retrieval of a new representation of a
    resource is undesirable for equivalent URIs, such as "file:///x" and
    "file://localhost/x" (see RFC 1738), or "http://spam/~x/",
    "http://spam/%7Ex/" and "http://spam/%7ex" (see RFC 3986).

    Normalization performed includes case normalization on the scheme and
    percent-encoded octets, percent-encoding normalization (decoding of
    octets corresponding to unreserved characters), and the reduction of
    'file://localhost/' to 'file:///', in accordance with both RFC 1738 and
    RFC 3986 (although RFC 3986 encourages using 'localhost' and doing
    this for all schemes, not just file).

    An instance of this class is used by Ft.Xml.Xslt.XsltContext for caching
    documents, so that the XSLT function document() will return identical
    nodes, without refetching/reparsing, for equivalent URIs.
    """
    # RFC 3986 requires localhost to be the default host no matter
    # what the scheme, but, being descriptive of existing practices,
    # leaves it up to the implementation to decide whether to use this
    # and other tests of URI equivalence in the determination of
    # same-document references. So our implementation results in what
    # is arguably desirable, but not strictly required, behavior.
    #
    #FIXME: make localhost the default for all schemes, not just file
    def _normalizekey(self, key):
        key = normalize_case(normalize_percent_encoding(key))
        if key[:17] == 'file://localhost/':
            return 'file://' + key[16:]
        else:
            return key

    def __getitem__(self, key):
        return super(uridict, self).__getitem__(self._normalizekey(key))

    def __setitem__(self, key, value):
        return super(uridict, self).__setitem__(self._normalizekey(key), value)

    def __delitem__(self, key):
        return super(uridict, self).__delitem__(self._normalizekey(key))

    def has_key(self, key):
        return super(uridict, self).has_key(self._normalizekey(key))

    def __contains__(self, key):
        return super(uridict, self).__contains__(self._normalizekey(key))

    def __iter__(self):
        return iter(self.keys())

    iterkeys = __iter__
    def iteritems(self):
        for key in self.iterkeys():
            yield key, self.__getitem__(key)


#FIXME: Port to more amara.lib.iri functions
def get_filename_from_url(url):
    fullname = url.split('/')[-1].split('#')[0].split('?')[0]
    return fullname


def get_filename_parts_from_url(url):
    fullname = url.split('/')[-1].split('#')[0].split('?')[0]
    t = list(os.path.splitext(fullname))
    if t[1]:
        t[1] = t[1][1:]
    return t


#Monkeypatch workaround to http://bugs.python.org/issue17214 
#("http.client.HTTPConnection.putrequest encode error")
#In order to use this, just do: 
#from amara3 import irihelper
#Then for example you can safely use stdlib urlopen or amara3.iri.urlopen in cases affected by encoding error after redirect

#from amara3.iri import urlopen

from urllib.request import HTTPRedirectHandler
# Implementation note: To avoid the server sending us into an
# infinite loop, the request object needs to track what URLs we
# have already seen.  Do this by adding a handler-specific
# attribute to the Request object.
def http_error_302(self, req, fp, code, msg, headers):
    # Some servers (incorrectly) return multiple Location headers
    # (so probably same goes for URI).  Use first header.
    if "location" in headers:
        newurl = headers["location"]
    elif "uri" in headers:
        newurl = headers["uri"]
    else:
        return

    # fix a possible malformed URL
    urlparts = urlparse(newurl)

    # For security reasons we don't allow redirection to anything other
    # than http, https or ftp.

    if urlparts.scheme not in ('http', 'https', 'ftp', ''):
        raise HTTPError(
            newurl, code,
            "%s - Redirection to url '%s' is not allowed" % (msg, newurl),
            headers, fp)

    urlpartslist = list(urlparts)
    path = urlparts.path if urlpaths.path else "/"
    # parse_headers() decodes from iso-8859-1 and unquotes, undo damage
    path = urlparts.path.encode("iso-8859-1")
    urlpartslist[2] = quote(path)

    newurl = urlunparse(urlpartslist)
    newurl = urljoin(req.full_url, newurl)

    # XXX Probably want to forget about the state of the current
    # request, although that might interact poorly with other
    # handlers that also use handler-specific request attributes
    new = self.redirect_request(req, fp, code, msg, headers, newurl)
    if new is None:
        return

    # loop detection
    # .redirect_dict has a key url if url was previously visited.
    if hasattr(req, 'redirect_dict'):
        visited = new.redirect_dict = req.redirect_dict
        if (visited.get(newurl, 0) >= self.max_repeats or
            len(visited) >= self.max_redirections):
            raise HTTPError(req.full_url, code,
                            self.inf_msg + msg, headers, fp)
    else:
        visited = new.redirect_dict = req.redirect_dict = {}
    visited[newurl] = visited.get(newurl, 0) + 1

    # Don't close the fp until we are sure that we won't use it
    # with HTTPError.
    fp.read()
    fp.close()

    return self.parent.open(new, timeout=req.timeout)

HTTPRedirectHandler.http_error_301 = HTTPRedirectHandler.http_error_303 = HTTPRedirectHandler.http_error_307 = http_error_302




# amara3.util
"""
Some utilities for general use in Amara

Copyright 2008-2015 Uche Ogbuji
"""

def coroutine(func):
    '''
    Decorator: Eliminate the need to call next() to kick-start a co-routine
    From David Beazley: http://www.dabeaz.com/generators/index.html
    '''
    def start(*args,**kwargs):
        coro = func(*args,**kwargs)
        next(coro)
        return coro
    return start


def parse_requirement(r):
    '''
    Parse out package & version from requirements.txt format
    
    For example, might use in setup.py as follows (for requires=REQUIREMENTS, in the setup function):
    
    with open(REQ_FILENAME) as infp:
        #See https://pip.pypa.io/en/stable/reference/pip_install/#requirements-file-format
        reqs = [r.split('#', 1)[0].strip() for r in infp.read().split('\n') if r.split('#', 1)[0].strip() ]
        REQUIREMENTS = [parse_requirement(r) for r in reqs]
    
    >>> from amara3.util import parse_requirement
    >>> reqtxt = """amara3-iri==3.0.0b3
    ... pymarc
    ... rdflib
    ... mmh3
    ... pytest
    ... versa>=0.3.3
    ... 
    ... #From Versa pyreqs.txt
    ... amara3-xml==3.0.0a6
    ... """
    >>> for line in reqtxt.splitlines():
    ...     parse_requirement(line)
    ... 
    'amara3 (==3.0.0b3)'
    'pymarc'
    'rdflib'
    'mmh3'
    'pytest'
    'versa (>=0.3.3)'
    ''
    '#From Versa pyreqs.txt'
    'amara3 (==3.0.0a6)'
    '''
    import re #Defying PEP 8 for perf of other routines
    m = re.search('[<>=][=]', r)
    if m:
        package = r[:m.start()]
        version = r[m.start():]
        if '-' in package:
            package = package.split('-')[0]
        return '{} ({})'.format(package, version)
    else:
        return r





"""
Classes and functions related to IRI/URI processing, validation, resolution, etc.

Copyright 2008-2018 Uche Ogbuji and Mike Brown
"""

__all__ = [
  "IriError",
  # IRI tools
  "iri_to_uri",
  "nfc_normalize",
  "convert_ireg_name",

  # RFC 3986 implementation
  'matches_uri_ref_syntax', 'matches_uri_syntax',
  'percent_encode', 'percent_decode',
  'split_uri_ref', 'unsplit_uri_ref',
  'split_authority', 'split_fragment',
  'absolutize', 'relativize', 'remove_dot_segments',
  'normalize_case', 'normalize_percent_encoding',
  'normalize_path_segments', 'normalize_path_segments_in_uri',

  # RFC 3151 implementation
  'urn_to_public_id', 'public_id_to_urn',

  # Miscellaneous
  'is_absolute', 'get_scheme', 'strip_fragment',
  'os_path_to_uri', 'uri_to_os_path', 'basejoin', 'join',
  'WINDOWS_SLASH_COMPAT', 'path_resolve',
]

import os, sys
import urllib, urllib.request
import re, io
import email
from string import ascii_letters
from email.utils import formatdate as _formatdate
from uuid import UUID, uuid1, uuid4

# whether os_path_to_uri should treat "/" same as "\" in a Windows path
WINDOWS_SLASH_COMPAT = True

DEFAULT_HIERARCHICAL_SEP = '/'

PERCENT_DECODE_BYTES = ('0123456789%s-._~' % ascii_letters).encode('ascii')

class IriError(Exception):
    """
    Exception related to URI/IRI processing
    """
    pass
    #FIXME: Re-incorporate the exception details below
'''
        return {
            IriError.INVALID_BASE_URI: _(
                "Invalid base URI: %(base)r cannot be used to resolve "
                " reference %(ref)r"),
            IriError.RELATIVE_BASE_URI: _(
                "Invalid base URI: %(base)r cannot be used to resolve "
                "reference %(ref)r; the base URI must be absolute, not "
                "relative."),
            IriError.NON_FILE_URI: _(
                "Only a 'file' URI can be converted to an OS-specific path; "
                "URI given was %(uri)r"),
            IriError.UNIX_REMOTE_HOST_FILE_URI: _(
                "A URI containing a remote host name cannot be converted to a "
                " path on posix; URI given was %(uri)r"),
            IriError.RESOURCE_ERROR: _(
                "Error retrieving resource %(loc)r: %(msg)s"),
            IriError.UNSUPPORTED_PLATFORM: _(
                "Platform %(platform)r not supported by URI function "
                "%(function)s"),
            IriError.SCHEME_REQUIRED: _(
                "Scheme-based resolution requires a URI with a scheme; "
                "neither the base URI %(base)r nor the reference %(ref)r "
                "have one."),
            IriError.INVALID_PUBLIC_ID_URN: _(
                "A public ID cannot be derived from URN %(urn)r "
                "because it does not conform to RFC 3151."),
            IriError.UNSUPPORTED_SCHEME: _(
                "The URI scheme %(scheme)s is not supported by resolver "),
            IriError.DENIED_BY_RULE: _(
                "Access to IRI %(uri)r was denied by action of an IRI restriction"),
            }
'''


def iri_to_uri(iri, convertHost=False):
    r"""
    Converts an IRI or IRI reference to a URI or URI reference,
    implementing sec. 3.1 of draft-duerst-iri-10.

    The convertHost flag indicates whether to perform conversion of
    the ireg-name (host) component of the IRI to an RFC 2396-compatible
    URI reg-name (IDNA encoded), e.g.
    iri_to_uri('http://r\xe9sum\xe9.example.org/', convertHost=False)
    => 'http://r%C3%A9sum%C3%A9.example.org/'
    iri_to_uri('http://r\xe9sum\xe9.example.org/', convertHost=True)
    => 'http://xn--rsum-bpad.example.org/'

    Ordinarily, the IRI should be given as a unicode string. If the IRI
    is instead given as a byte string, then it will be assumed to be
    UTF-8 encoded, will be decoded accordingly, and as per the
    requirements of the conversion algorithm, will NOT be normalized.
    """
    if not isinstance(iri, str):
        iri = nfc_normalize(iri)

    # first we have to get the host
    (scheme, auth, path, query, frag) = split_uri_ref(iri)
    if auth and auth.find('@') > -1:
        userinfo, hostport = auth.split('@')
    else:
        userinfo = None
        hostport = auth
    if hostport and hostport.find(':') > -1:
        host, port = hostport.split(':')
    else:
        host = hostport
        port = None
    if host:
        host = convert_ireg_name(host)
        auth = ''
        if userinfo:
            auth += userinfo + '@'
        auth += host
        if port:
            auth += ':' + port
    iri = unsplit_uri_ref((scheme, auth, path, query, frag))

    res = ''
    pos = 0
    #FIXME: use re.subn with substitution function for big speed-up
    surrogate = None
    for c in iri:
        cp = ord(c)
        if cp > 128:
            if cp < 160:
                # FIXME: i18n
                raise ValueError(_("Illegal character at position %d (0-based) of IRI %r" % (pos, iri)))
            # 'for c in iri' may give us surrogate pairs
            elif cp > 55295:
                if cp < 56320:
                    # d800-dbff
                    surrogate = c
                    continue
                elif cp < 57344:
                    # dc00-dfff
                    if surrogate is None:
                        raise ValueError(_("Illegal surrogate pair in %r" % iri))
                    c = surrogate + c
                else:
                    raise ValueError(_("Illegal surrogate pair in %r" % iri))
                surrogate = None
            for octet in c.encode('utf-8'):
                res += '%%%02X' % ord(octet)
        else:
            res += c
        pos += 1
    return res


def nfc_normalize(iri):
    """
    Normalizes the given unicode string according to Unicode Normalization Form C (NFC)
    so that it can be used as an IRI or IRI reference.
    """
    from unicodedata import normalize
    return normalize('NFC', iri)


def convert_ireg_name(iregname):
    """
    Converts the given ireg-name component of an IRI to a string suitable for use
    as a URI reg-name in pre-rfc2396bis schemes and resolvers. Returns the ireg-name
    """
    # I have not yet verified that the default IDNA encoding
    # matches the algorithm required by the IRI spec, but it
    # does work on the one simple example in the spec.
    return iregname.encode('idna').decode('ascii')


#=============================================================================
# Functions that implement aspects of RFC 3986
#
_validation_setup_completed = False
def _init_uri_validation_regex():
    """
    Called internally to compile the regular expressions needed by
    URI validation functions, just once, the first time a function
    that needs them is called.
    """
    global _validation_setup_completed
    if _validation_setup_completed:
        return

    #-------------------------------------------------------------------------
    # Regular expressions for determining the non-URI-ness of strings
    #
    # A given string's designation as a URI or URI reference comes from the
    # context in which it is being used, not from its syntax; a regular
    # expression can at most only determine whether a given string COULD be a
    # URI or URI reference, based on its lexical structure.
    #
    # 1. Altova's regex (in the public domain; courtesy Altova)
    #
    # # based on the BNF grammar in the original RFC 2396
    # ALTOVA_REGEX = r"(([a-zA-Z][0-9a-zA-Z+\-\.]*:)?/{0,2}" + \
    #                r"[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?" + \
    #                r"(#[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?"
    #
    # This regex matches URI references, and thus URIs as well. It is also
    # lenient; some strings that are not URI references can falsely match.
    #
    # It is also not very useful as-is, because it essentially has the form
    # (group1)?(group2)? -- this matches the empty string, and in fact any
    # string or substring can be said to match this pattern. To be useful,
    # this regex (and any like it) must be changed so that it only matches
    # an entire string. This is accomplished in Python by using the \A and \Z
    # delimiters around the pattern:
    #
    # BETTER_ALTOVA_REGEX = r"\A(?!\n)%s\Z" % ALTOVA_REGEX
    #
    # The (?!\n) takes care of an edge case where a string consisting of a
    # sole linefeed character would falsely match.
    #
    # 2. Python regular expressions for strict validation of URIs and URI
    #    references (in the public domain; courtesy Fourthought, Inc.)
    #
    # Note that we do not use any \d or \w shortcuts, as these are
    # potentially locale or Unicode sensitive.
    #
    # # based on the ABNF in RFC 3986,
    # # "Uniform Resource Identifier (URI): Generic Syntax"
    pchar           = r"(?:[0-9A-Za-z\-_\.!~*'();:@&=+$,]|(?:%[0-9A-Fa-f]{2}))"
    fragment        = r"(?:[0-9A-Za-z\-_\.!~*'();:@&=+$,/?]|(?:%[0-9A-Fa-f]{2}))*"
    query           = fragment
    segment_nz_nc   = r"(?:[0-9A-Za-z\-_\.!~*'();@&=+$,]|(?:%[0-9A-Fa-f]{2}))+"
    segment_nz      = r'%s+' % pchar
    segment         = r'%s*' % pchar
    #path_empty      = r''  # zero characters
    path_rootless   = r'%s(?:/%s)*' % (segment_nz, segment)   # begins with a segment
    path_noscheme   = r'%s(?:/%s)*' % (segment_nz_nc, segment)  # begins with a non-colon segment
    path_absolute   = r'/(?:%s)?' % path_rootless  # begins with "/" but not "//"
    path_abempty    = r'(?:/%s)*' % segment   # begins with "/" or is empty
    #path            = r'(?:(?:%s)|(?:%s)|(?:%s)|(?:%s))?' % (path_abempty, path_absolute, path_noscheme, path_rootless)
    domainlabel     = r'[0-9A-Za-z](?:[0-9A-Za-z\-]{0,61}[0-9A-Za-z])?'
    qualified       = r'(?:\.%s)*\.?' % domainlabel
    reg_name        = r"(?:(?:[0-9A-Za-z\-_\.!~*'();&=+$,]|(?:%[0-9A-Fa-f]{2}))*)"
    dec_octet       = r'(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])'
    IPv4address     = r'(?:%s\.){3}(?:%s)' % (dec_octet, dec_octet)
    h16             = r'[0-9A-Fa-f]{1,4}'
    ls32            = r'(?:(?:%s:%s)|%s)' % (h16, h16, IPv4address)
    IPv6address     = r'(?:' + \
                      r'(?:(?:%s:){6}%s)' % (h16, ls32) + \
                      r'|(?:::(?:%s:){5}%s)' % (h16, ls32) + \
                      r'|(?:%s?::(?:%s:){4}%s)' % (h16, h16, ls32) + \
                      r'|(?:(?:(?:%s:)?%s)?::(?:%s:){3}%s)' % (h16, h16, h16, ls32) + \
                      r'|(?:(?:(?:%s:)?%s){0,2}::(?:%s:){2}%s)' % (h16, h16, h16, ls32) + \
                      r'|(?:(?:(?:%s:)?%s){0,3}::%s:%s)' % (h16, h16, h16, ls32) + \
                      r'|(?:(?:(?:%s:)?%s){0,4}::%s)' % (h16, h16, ls32) + \
                      r'|(?:(?:(?:%s:)?%s){0,5}::%s)' % (h16, h16, h16) + \
                      r'|(?:(?:(?:%s:)?%s){0,6}::)' % (h16, h16) + \
                      r')'
    IPvFuture       = r"(?:v[0-9A-Fa-f]+\.[0-9A-Za-z\-\._~!$&'()*+,;=:]+)"
    IP_literal      = r'\[(?:%s|%s)\]' % (IPv6address, IPvFuture)
    port            = r'[0-9]*'
    host            = r'(?:%s|%s|%s)?' % (IP_literal, IPv4address, reg_name)
    userinfo        = r"(?:[0-9A-Za-z\-_\.!~*'();:@&=+$,]|(?:%[0-9A-Fa-f]{2}))*"
    authority       = r'(?:%s@)?%s(?::%s)?' % (userinfo, host, port)
    scheme          = r'[A-Za-z][0-9A-Za-z+\-\.]*'
    #absolute_URI    = r'%s:%s(?:\?%s)?' % (scheme, hier_part, query)
    relative_part   = r'(?:(?://%s%s)|(?:%s)|(?:%s))?' % (authority, path_abempty,
                                                          path_absolute, path_noscheme)
    relative_ref    = r'%s(?:\?%s)?(?:#%s)?' % (relative_part, query, fragment)
    hier_part       = r'(?:(?://%s%s)|(?:%s)|(?:%s))?' % (authority, path_abempty,
                                                          path_absolute, path_rootless)
    URI             = r'%s:%s(?:\?%s)?(?:#%s)?' % (scheme, hier_part, query, fragment)
    URI_reference   = r'(?:%s|%s)' % (URI, relative_ref)

    STRICT_URI_PYREGEX = r"\A%s\Z" % URI
    STRICT_URIREF_PYREGEX = r"\A(?!\n)%s\Z" % URI_reference

    global URI_PATTERN, URI_REF_PATTERN
    URI_PATTERN = re.compile(STRICT_URI_PYREGEX)        # strict checking for URIs
    URI_REF_PATTERN = re.compile(STRICT_URIREF_PYREGEX) # strict checking for URI refs
    _validation_setup_completed = True
    return


def matches_uri_ref_syntax(s):
    """
    This function returns true if the given string could be a URI reference,
    as defined in RFC 3986, just based on the string's syntax.

    A URI reference can be a URI or certain portions of one, including the
    empty string, and it can have a fragment component.
    """
    if not _validation_setup_completed:
        _init_uri_validation_regex()
    return URI_REF_PATTERN.match(s) is not None


def matches_uri_syntax(s):
    """
    This function returns true if the given string could be a URI, as defined
    in RFC 3986, just based on the string's syntax.

    A URI is by definition absolute (begins with a scheme) and does not end
    with a #fragment. It also must adhere to various other syntax rules.
    """
    if not _validation_setup_completed:
        _init_uri_validation_regex()
    return URI_PATTERN.match(s) is not None


_split_uri_ref_setup_completed = False
def _init_split_uri_ref_pattern():
    """
    Called internally to compile the regular expression used by
    split_uri_ref() just once, the first time the function is called.
    """
    global _split_uri_ref_setup_completed
    if _split_uri_ref_setup_completed:
        return

    # Like the others, this regex is also in the public domain.
    # It is based on this one, from RFC 3986 appendix B
    # (unchanged from RFC 2396 appendix B):
    # ^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?
    regex = r"^(?:(?P<scheme>[^:/?#]+):)?(?://(?P<authority>[^/?#]*))?(?P<path>[^?#]*)(?:\?(?P<query>[^#]*))?(?:#(?P<fragment>.*))?$"
    global SPLIT_URI_REF_PATTERN
    SPLIT_URI_REF_PATTERN = re.compile(regex)
    _split_uri_ref_setup_completed = True
    return


def split_uri_ref(uriref):
    """
    Given a valid URI reference as a string, returns a tuple representing the
    generic URI components, as per RFC 3986 appendix B. The tuple's structure
    is (scheme, authority, path, query, fragment).

    All values will be strings (possibly empty) or None if undefined.

    Note that per RFC 3986, there is no distinction between a path and
    an "opaque part", as there was in RFC 2396.
    """
    if not _split_uri_ref_setup_completed:
        _init_split_uri_ref_pattern()
    # the pattern will match every possible string, so it's safe to
    # assume there's a groupdict method to call.
    g = SPLIT_URI_REF_PATTERN.match(uriref).groupdict()
    scheme      = g['scheme']
    authority   = g['authority']
    path        = g['path']
    query       = g['query']
    fragment    = g['fragment']
    return (scheme, authority, path, query, fragment)


def unsplit_uri_ref(uriRefSeq):
    """
    Given a sequence as would be produced by split_uri_ref(), assembles and
    returns a URI reference as a string.
    """
    if not isinstance(uriRefSeq, (tuple, list)):
        raise TypeError(_("sequence expected, got %s" % type(uriRefSeq)))
    (scheme, authority, path, query, fragment) = uriRefSeq
    uri = ''
    if scheme is not None:
        uri += scheme + ':'
    if authority is not None:
        uri += '//' + authority
    uri += path
    if query is not None:
        uri += '?' + query
    if fragment is not None:
        uri += '#' + fragment
    return uri


_split_authority_setup_completed = False
def _init_split_authority_pattern():
    """
    Called internally to compile the regular expression used by
    split_authority() just once, the first time the function is called.
    """
    global _split_authority_setup_completed
    if _split_authority_setup_completed:
        return
    global SPLIT_AUTHORITY_PATTERN
    regex = r'(?:(?P<userinfo>[^@]*)@)?(?P<host>[^:]*)(?::(?P<port>.*))?'
    SPLIT_AUTHORITY_PATTERN = re.compile(regex)
    _split_authority_setup_completed = True
    return


def split_authority(authority):
    """
    Given a string representing the authority component of a URI, returns
    a tuple consisting of the subcomponents (userinfo, host, port). No
    percent-decoding is performed.
    """
    if not _split_authority_setup_completed:
        _init_split_authority_pattern()
    m = SPLIT_AUTHORITY_PATTERN.match(authority)
    if m:
        return m.groups()
    else:
        return (None, authority, None)


def split_fragment(uri):
    """
    Given a URI or URI reference, returns a tuple consisting of
    (base, fragment), where base is the portion before the '#' that
    precedes the fragment component.
    """
    # The only '#' in a legit URI will be the fragment separator,
    # but in the wild, people get sloppy. Assume the last '#' is it.
    pos = uri.rfind('#')
    if pos == -1:
        return (uri, uri[:0])
    else:
        return (uri[:pos], uri[pos+1:])


# "unreserved" characters are allowed in a URI, and do not have special
# meaning as delimiters of URI components or subcomponents. They may
# appear raw or percent-encoded, but percent-encoding is discouraged.
# This set of characters is sufficiently long enough that using a
# compiled regex is faster than using a string with the "in" operator.
#UNRESERVED_PATTERN = re.compile(r"[0-9A-Za-z\-\._~!*'()]") # RFC 2396
UNRESERVED_PATTERN = re.compile(r'[0-9A-Za-z\-\._~]') # RFC 3986

# "reserved" characters are allowed in a URI, but they may or always do
# have special meaning as delimiters of URI components or subcomponents.
# When being used as delimiters, they must be raw, and when not being
# used as delimiters, they must be percent-encoded.
# This set of characters is sufficiently short enough that using a
# string with the "in" operator is faster than using a compiled regex.
# The characters in the string are ordered according to how likely they
# are to be found (approximately), for faster operation with "in".
#RESERVED = "/&=+?;@,:$[]" # RFC 2396 + RFC 2732
RESERVED = "/=&+?#;@,:$!*[]()'" # RFC 3986


def percent_encode(s, encoding='utf-8', encodeReserved=True, spaceToPlus=False,
                     nlChars=None, reservedChars=RESERVED):
    """
    [*** Experimental API ***] This function applies percent-encoding, as
    described in RFC 3986 sec. 2.1, to the given string, in order to prepare
    the string for use in a URI. It replaces characters that are not allowed
    in a URI. By default, it also replaces characters in the reserved set,
    which normally includes the generic URI component delimiters ":" "/"
    "?" \"#\" "[" "]" "@" and the subcomponent delimiters "!" "$" "&" "\'" "("
    ")" "*" "+" "," ";" "=".

    Ideally, this function should be used on individual components or
    subcomponents of a URI prior to assembly of the complete URI, not
    afterward, because this function has no way of knowing which characters
    in the reserved set are being used for their reserved purpose and which
    are part of the data. By default it assumes that they are all being used
    as data, thus they all become percent-encoded.

    The characters in the reserved set can be overridden from the default by
    setting the reservedChars argument. The percent-encoding of characters
    in the reserved set can be disabled by unsetting the encodeReserved flag.
    Do this if the string is an already-assembled URI or a URI component,
    such as a complete path.

    The encoding argument will be used to determine the percent-encoded octets
    for characters that are not in the U+0000 to U+007F range. The codec
    identified by the encoding argument must return a byte string.

    The spaceToPlus flag controls whether space characters are changed to
    "+" characters in the result, rather than being percent-encoded.
    Generally, this is not required, and given the status of "+" as a
    reserved character, is often undesirable. But it is required in certain
    situations, such as when generating application/x-www-form-urlencoded
    content or RFC 3151 public identifier URNs, so it is supported here.

    The nlChars argument, if given, is a sequence type in which each member
    is a substring that indicates a "new line". Occurrences of this substring
    will be replaced by '%0D%0A' in the result, as is required when generating
    application/x-www-form-urlencoded content.

    This function is similar to urllib.quote(), but is more conformant and
    Unicode-friendly. Suggestions for improvements welcome.

    >>> from amara3 import iri
    >>> iri.percent_encode('http://bibfra.me/vocab/relation/論定')
    http%3A%2F%2Fbibfra.me%2Fvocab%2Frelation%2F%E8%AB%96%E5%AE%9A
    """
    res = ''
    if nlChars is not None:
        for c in nlChars:
            s.replace(c, '\r\n')
    #FIXME: use re.subn with substitution function for big speed-up
    for c in s:
        # surrogates? -> percent-encode according to given encoding
        if UNRESERVED_PATTERN.match(c) is None:
            cp = ord(c)
            # ASCII range?
            if cp < 128:
                # space? -> plus if desired
                if spaceToPlus and c == ' ':
                    res += '+'
                # reserved? -> percent-encode if desired
                elif c in reservedChars:
                    if encodeReserved:
                        res += '%%%02X' % cp
                    else:
                        res += c
                # not unreserved or reserved, so percent-encode
                # FIXME: should percent-encode according to given encoding;
                # ASCII range is not special!
                else:
                    res += '%%%02X' % cp
            # non-ASCII-range unicode?
            else:
                # percent-encode according to given encoding
                for octet in c.encode(encoding):
                    res += '%%%02X' % octet

        # unreserved -> safe to use as-is
        else:
            res += c
    return res

_ASCII_PAT = re.compile('([\x00-\x7f]+)')

_HEXDIG = '0123456789ABCDEFabcdef'
_HEXTOBYTE = None


def _unquote_to_bytes(s, decodable=None):
    """_unquote_to_bytes('abc%20def') -> b'abc def'."""
    # Note: strings are encoded as UTF-8. This is only an issue if it contains
    # unescaped non-ASCII characters, which URIs should not.
    if not s:
        # Is it a string-like object?
        s.split
        return b''
    if isinstance(s, str):
        s = s.encode('utf-8')
    bits = s.split(b'%')
    if len(bits) == 1:
        return s
    res = [bits[0]]
    append = res.append
    # Delay the initialization of the table to not waste memory
    # if the function is never called
    global _HEXTOBYTE
    if _HEXTOBYTE is None:
        _HEXTOBYTE = {(a + b).encode(): bytes([int(a + b, 16)])
                      for a in _HEXDIG for b in _HEXDIG}
    for item in bits[1:]:
        try:
            c = chr(int(item[:2], 16)).encode('ascii')
            if decodable is None or c in decodable:
                append(_HEXTOBYTE[item[:2]])
                append(item[2:])
            #FIXME: We'll need to do our own surrogate pair decoding because:
            #>>> '\ud800'.encode('utf-8') -> UnicodeEncodeError: 'utf-8' codec can't encode character '\ud800' in position 0: surrogates not allowed
            else:
                append(b'%')
                append(item)
        except (ValueError, KeyError):
            append(b'%')
            append(item)
    return b''.join(res)

#>>> from amara3.iri import percent_decode
#>>> u0 = 'example://A/b/c/%7bfoo%7d'
#>>> u1 = percent_decode(u0)
#>>> u1
#'example://A/b/c/{foo}'


def percent_decode(s, encoding='utf-8', decodable=None, errors='replace'):
    """
    [*** Experimental API ***] Reverses the percent-encoding of the given
    string.

    Similar to urllib.parse.unquote()

    By default, all percent-encoded sequences are decoded, but if a byte
    string is given via the 'decodable' argument, only the sequences
    corresponding to those octets will be decoded.

    Percent-encoded sequences are converted to bytes, then converted back to
    string (Unicode) according to the given encoding.
    For example, by default, 'abc%E2%80%A2' will be converted to 'abc\u2022',
    because byte sequence E2 80 A2 represents character U+2022 in UTF-8.

    This function is intended for use on the portions of a URI that are
    delimited by reserved characters (see percent_encode), or on a value from
    data of media type application/x-www-form-urlencoded.

    >>> from amara3.iri import percent_decode
    >>> u0 = 'http://host/abc%E2%80%A2/x/y/z'
    >>> u1 = percent_decode(u0)
    >>> hex(ord(u1[15]))
    '0x2022'
    """
    # Most of this comes from urllib.parse.unquote().
    # Note: strings are encoded as UTF-8. This is only an issue if it contains
    # unescaped non-ASCII characters, which URIs should not.
    # If given a string argument, does not decode
    # percent-encoded octets above %7F.

    if '%' not in s:
        #s.split
        return s
    if encoding is None:
        encoding = 'utf-8'
    if errors is None:
        errors = 'replace'
    bits = _ASCII_PAT.split(s)
    res = [bits[0]]
    append = res.append #Saving the func lookups in the tight loop below
    for i in range(1, len(bits), 2):
        append(_unquote_to_bytes(bits[i], decodable=decodable).decode(encoding, errors))
        append(bits[i + 1])
    return ''.join(res)


def absolutize(uriRef, baseUri, limit_schemes=None):
    """
    Resolves a URI reference to absolute form, effecting the result of RFC
    3986 section 5. The URI reference is considered to be relative to the
    given base URI.

    It is the caller's responsibility to ensure that the base URI matches
    the absolute-URI syntax rule of RFC 3986, and that its path component
    does not contain '.' or '..' segments if the scheme is hierarchical.
    Unexpected results may occur otherwise.

    This function only conducts a minimal sanity check in order to determine
    if relative resolution is possible: it raises a ValueError if the base
    URI does not have a scheme component. While it is true that the base URI
    is irrelevant if the URI reference has a scheme, an exception is raised
    in order to signal that the given string does not even come close to
    meeting the criteria to be usable as a base URI.

    It is the caller's responsibility to make a determination of whether the
    URI reference constitutes a "same-document reference", as defined in RFC
    2396 or RFC 3986. As per the spec, dereferencing a same-document
    reference "should not" involve retrieval of a new representation of the
    referenced resource. Note that the two specs have different definitions
    of same-document reference: RFC 2396 says it is *only* the cases where the
    reference is the empty string, or \"#\" followed by a fragment; RFC 3986
    requires making a comparison of the base URI to the absolute form of the
    reference (as is returned by the spec), minus its fragment component,
    if any.

    This function is similar to urlparse.urljoin() and urllib.basejoin().
    Those functions, however, are (as of Python 2.3) outdated, buggy, and/or
    designed to produce results acceptable for use with other core Python
    libraries, rather than being earnest implementations of the relevant
    specs. Their problems are most noticeable in their handling of
    same-document references and 'file:' URIs, both being situations that
    come up far too often to consider the functions reliable enough for
    general use.
    """
    # Reasons to avoid using urllib.basejoin() and urlparse.urljoin():
    # - Both are partial implementations of long-obsolete specs.
    # - Both accept relative URLs as the base, which no spec allows.
    # - urllib.basejoin() mishandles the '' and '..' references.
    # - If the base URL uses a non-hierarchical or relative path,
    #    or if the URL scheme is unrecognized, the result is not
    #    always as expected (partly due to issues in RFC 1808).
    # - If the authority component of a 'file' URI is empty,
    #    the authority component is removed altogether. If it was
    #    not present, an empty authority component is in the result.
    # - '.' and '..' segments are not always collapsed as well as they
    #    should be (partly due to issues in RFC 1808).
    # - Effective Python 2.4, urllib.basejoin() *is* urlparse.urljoin(),
    #    but urlparse.urljoin() is still based on RFC 1808.

    # This procedure is based on the pseudocode in RFC 3986 sec. 5.2.
    #
    # ensure base URI is absolute
    if is_absolute(uriRef):
        return uriRef
    if not baseUri or not is_absolute(baseUri):
        raise ValueError("Invalid base URI: {base} cannot be used to resolve "
                "reference {ref}; the base URI must be absolute, not "
                "relative.".format(base=baseUri, ref=uriRef))
    if limit_schemes and get_scheme(baseUri) not in limit_schemes:
        scheme = get_scheme(baseUri)
        raise ValueError("The URI scheme {scheme} is not supported by resolver".format(scheme=scheme))

    # shortcut for the simplest same-document reference cases
    if uriRef == '' or uriRef[0] == '#':
        return baseUri.split('#')[0] + uriRef
    # ensure a clean slate
    tScheme = tAuth = tPath = tQuery = None
    # parse the reference into its components
    (rScheme, rAuth, rPath, rQuery, rFrag) = split_uri_ref(uriRef)
    # if the reference is absolute, eliminate '.' and '..' path segments
    # and skip to the end
    if rScheme is not None:
        tScheme = rScheme
        tAuth = rAuth
        tPath = remove_dot_segments(rPath)
        tQuery = rQuery
    else:
        # the base URI's scheme, and possibly more, will be inherited
        (bScheme, bAuth, bPath, bQuery, bFrag) = split_uri_ref(baseUri)
        # if the reference is a net-path, just eliminate '.' and '..' path
        # segments; no other changes needed.
        if rAuth is not None:
            tAuth = rAuth
            tPath = remove_dot_segments(rPath)
            tQuery = rQuery
        # if it's not a net-path, we need to inherit pieces of the base URI
        else:
            # use base URI's path if the reference's path is empty
            if not rPath:
                tPath = bPath
                # use the reference's query, if any, or else the base URI's,
                tQuery = rQuery is not None and rQuery or bQuery
            # the reference's path is not empty
            else:
                # just use the reference's path if it's absolute
                if rPath[0] == '/':
                    tPath = remove_dot_segments(rPath)
                # merge the reference's relative path with the base URI's path
                else:
                    if bAuth is not None and not bPath:
                        tPath = '/' + rPath
                    else:
                        tPath = bPath[:bPath.rfind('/')+1] + rPath
                    tPath = remove_dot_segments(tPath)
                # use the reference's query
                tQuery = rQuery
            # since the reference isn't a net-path,
            # use the authority from the base URI
            tAuth = bAuth
        # inherit the scheme from the base URI
        tScheme = bScheme
    # always use the reference's fragment (but no need to define another var)
    #tFrag = rFrag

    # now compose the target URI (RFC 3986 sec. 5.3)
    return unsplit_uri_ref((tScheme, tAuth, tPath, tQuery, rFrag))


def relativize(targetUri, againstUri, subPathOnly=False):
    """
    This method returns a relative URI that is consistent with `targetURI`
    when resolved against `againstUri`.  If no such relative URI exists, for
    whatever reason, this method returns `None`.

    To be precise, if a string called `rel` exists such that
    ``absolutize(rel, againstUri) == targetUri``, then `rel` is returned by
    this function.  In these cases, `relativize` is in a sense the inverse
    of `absolutize`.  In all other cases, `relativize` returns `None`.

    The following idiom may be useful for obtaining compliant relative
    reference strings (e.g. for `path`) for use in other methods of this
    package::

      path = relativize(os_path_to_uri(path), os_path_to_uri('.'))

    If `subPathOnly` is `True`, then this method will only return a relative
    reference if such a reference exists relative to the last hierarchical
    segment of `againstUri`.  In particular, this relative reference will
    not start with '/' or '../'.
    """

    # We might want to change the semantics slightly to allow a relative
    # target URI to be a valid "relative path" (and just return it).  For
    # now, though, absolute URIs only.
    if not is_absolute(targetUri) or not is_absolute(againstUri):
        return None

    targetUri = normalize_path_segments_in_uri(targetUri)
    againstUri = normalize_path_segments_in_uri(againstUri)

    splitTarget = list(split_uri_ref(absolutize(targetUri, targetUri)))
    splitAgainst = list(split_uri_ref(absolutize(againstUri, againstUri)))

    if not splitTarget[:2] == splitAgainst[:2]:
        return None

    subPathSplit = [None, None] + splitTarget[2:]

    targetPath = splitTarget[2]
    againstPath = splitAgainst[2] or '/'

    leadingSlash = False
    if targetPath[:1] == '/' or againstPath[:1] == '/':
        if targetPath[:1] == againstPath[:1]:
            targetPath = targetPath[1:]
            againstPath = againstPath[1:]
            leadingSlash = True
        else:
            return None

    targetPathSegments = targetPath.split('/')
    againstPathSegments = againstPath.split('/')

    # Count the number of path segments in common.
    i = 0
    while True:
        # Stop if we get to the end of either segment list.
        if not(len(targetPathSegments) > i and
               len(againstPathSegments) > i):
            break

        # Increment the count when the lists agree, unless we are at the
        # last segment of either list and that segment is an empty segment.
        # We bail on this case because an empty ending segment in one path
        # must not match a mid-path empty segment in the other.
        if (targetPathSegments[i] == againstPathSegments[i]
            and not (i + 1 == len(againstPathSegments) and
                     '' == againstPathSegments[i])
            and not (i + 1 == len(targetPathSegments) and
                     '' == targetPathSegments[i])):
            i = i + 1
        # Otherwise stop.
        else:
            break

    # The target path has `i` segments in common with the basis path, and
    # the last segment (after the final '/') doesn't matter; we'll need to
    # traverse the rest.
    traverse = len(againstPathSegments) - i - 1

    relativePath = None
    # If the two paths do not agree on any segments, we have two special
    # cases.
    if i == 0 and leadingSlash:
        # First, if the ruling path only had one segment, then our result
        # can be a relative path.
        if len(againstPathSegments) == 1:
            relativePath = targetPath
        # Otherwise, the ruling path had a number of segments, so our result
        # must be an absolute path (unless we only want a subpath result, in
        # which case none exists).
        elif subPathOnly:
            return None
        else:
            relativePath = '/' + targetPath
    elif traverse > 0:
        if subPathOnly:
            return None
        relativePath = (("../" * traverse) +
                        '/'.join(targetPathSegments[i:]))
    # If the ith segment of the target path is empty and that is not the
    # final segment, then we need to precede the path with "./" to make it a
    # relative path.
    elif (len(targetPathSegments) > i + 1 and
          '' == targetPathSegments[i]):
        relativePath = "./" + '/'.join(targetPathSegments[i:])
    else:
        relativePath = '/'.join(targetPathSegments[i:])

    return unsplit_uri_ref([None, None, relativePath] + splitTarget[3:])


def remove_dot_segments(path):
    """
    Supports absolutize() by implementing the remove_dot_segments function
    described in RFC 3986 sec. 5.2.  It collapses most of the '.' and '..'
    segments out of a path without eliminating empty segments. It is intended
    to be used during the path merging process and may not give expected
    results when used independently. Use normalize_path_segments() or
    normalize_path_segments_in_uri() if more general normalization is desired.
    """
    # return empty string if entire path is just "." or ".."
    if path == '.' or path == '..':
        return path[0:0] # preserves string type
    # remove all "./" or "../" segments at the beginning
    while path:
        if path[:2] == './':
            path = path[2:]
        elif path[:3] == '../':
            path = path[3:]
        else:
            break
    # We need to keep track of whether there was a leading slash,
    # because we're going to drop it in order to prevent our list of
    # segments from having an ambiguous empty first item when we call
    # split().
    leading_slash = False
    if path[:1] == '/':
        path = path[1:]
        leading_slash = True
    # replace a trailing "/." with just "/"
    if path[-2:] == '/.':
        path = path[:-1]
    # convert the segments into a list and process each segment in
    # order from left to right.
    segments = path.split('/')
    keepers = []
    segments.reverse()
    while segments:
        seg = segments.pop()
        # '..' means drop the previous kept segment, if any.
        # If none, and if the path is relative, then keep the '..'.
        # If the '..' was the last segment, ensure
        # that the result ends with '/'.
        if seg == '..':
            if keepers:
                keepers.pop()
            elif not leading_slash:
                keepers.append(seg)
            if not segments:
                keepers.append('')
        # ignore '.' segments and keep all others, even empty ones
        elif seg != '.':
            keepers.append(seg)
    # reassemble the kept segments
    return leading_slash * '/' + '/'.join(keepers)


def normalize_case(uriRef, doHost=False):
    """
    Returns the given URI reference with the case of the scheme,
    percent-encoded octets, and, optionally, the host, all normalized,
    implementing section 6.2.2.1 of RFC 3986. The normal form of
    scheme and host is lowercase, and the normal form of
    percent-encoded octets is uppercase.

    The URI reference can be given as either a string or as a sequence as
    would be provided by the split_uri_ref function. The return value will
    be a string or tuple.
    """
    if not isinstance(uriRef, (tuple, list)):
        uriRef = split_uri_ref(uriRef)
        tup = None
    else:
        tup = True
    # normalize percent-encoded octets
    newRef = []
    for component in uriRef:
        if component:
            newRef.append(re.sub('%([0-9a-f][0-9a-f])',
                          lambda m: m.group(0).upper(), component))
        else:
            newRef.append(component)
    # normalize scheme
    scheme = newRef[0]
    if scheme:
        scheme = scheme.lower()
    # normalize host
    authority = newRef[1]
    if doHost:
        if authority:
            userinfo, host, port = split_authority(authority)
            authority = ''
            if userinfo is not None:
                authority += '%s@' % userinfo
            authority += host.lower()
            if port is not None:
                authority += ':%s' % port

    res = (scheme, authority, newRef[2], newRef[3], newRef[4])
    if tup:
        return res
    else:
        return unsplit_uri_ref(res)


def normalize_percent_encoding(s):
    """
    Given a string representing a URI reference or a component thereof,
    returns the string with all percent-encoded octets that correspond to
    unreserved characters decoded, implementing section 6.2.2.2 of RFC
    3986.

    >>> u0 = 'http://host/abc%E2%80%A2/x/y/z'
    >>> u1 = normalize_percent_encoding(u0)
    >>> hex(ord(u1[15]))
    '0x2022'
    """
    return percent_decode(s, decodable=PERCENT_DECODE_BYTES)


def normalize_path_segments(path):
    """
    Given a string representing the path component of a URI reference having a
    hierarchical scheme, returns the string with dot segments ('.' and '..')
    removed, implementing section 6.2.2.3 of RFC 3986. If the path is
    relative, it is returned with no changes.
    """
    if not path or path[:1] != '/':
        return path
    else:
        return remove_dot_segments(path)


def normalize_path_segments_in_uri(uri):
    """
    Given a string representing a URI or URI reference having a hierarchical
    scheme, returns the string with dot segments ('.' and '..') removed from
    the path component, implementing section 6.2.2.3 of RFC 3986. If the
    path is relative, the URI or URI reference is returned with no changes.
    """
    components = list(split_uri_ref(uri))
    components[2] = normalize_path_segments(components[2])
    return unsplit_uri_ref(components)


#=============================================================================
# RFC 3151 implementation
#

def urn_to_public_id(urn):
    """
    Converts a URN that conforms to RFC 3151 to a public identifier.

    For example, the URN
    "urn:publicid:%2B:IDN+example.org:DTD+XML+Bookmarks+1.0:EN:XML"
    will be converted to the public identifier
    "+//IDN example.org//DTD XML Bookmarks 1.0//EN//XML"

    Raises a ValueError if the given URN cannot be converted.
    Query and fragment components, if present, are ignored.
    """
    if urn is not None and urn:
        (scheme, auth, path, query, frag) = split_uri_ref(urn)
        if scheme is not None and scheme.lower() == 'urn':
            pp = path.split(':', 1)
            if len(pp) > 1:
                urn_scheme = percent_decode(pp[0])
                if urn_scheme == 'publicid':
                    publicid = pp[1].replace('+', ' ')
                    publicid = publicid.replace(':', '//')
                    publicid = publicid.replace(';', '::')
                    publicid = percent_decode(publicid)
                    return publicid

    raise ValueError("A public ID cannot be derived from URN {urn} "
                "because it does not conform to RFC 3151.".format(urn=urn))


def public_id_to_urn(publicid):
    """
    Converts a public identifier to a URN that conforms to RFC 3151.
    """
    # 1. condense whitespace, XSLT-style
    publicid = re.sub('[ \t\r\n]+', ' ', publicid.strip())
    # 2. // -> :
    #    :: -> ;
    #    space -> +
    #    + ; ' ? # % / : -> percent-encode
    #    (actually, the intent of the RFC is to not conflict with RFC 2396,
    #     so any character not in the unreserved set must be percent-encoded)
    r = ':'.join([';'.join([percent_encode(dcpart, spaceToPlus=True)
                            for dcpart in dspart.split('::')])
                  for dspart in publicid.split('//')])
    return 'urn:publicid:%s' % r


#=============================================================================
# Miscellaneous public functions
#

SCHEME_PATTERN = re.compile(r'([a-zA-Z][a-zA-Z0-9+\-.]*):')
def get_scheme(uriRef):
    """
    Obtains, with optimum efficiency, just the scheme from a URI reference.
    Returns a string, or if no scheme could be found, returns None.
    """
    # Using a regex seems to be the best option. Called 50,000 times on
    # different URIs, on a 1.0-GHz PIII with FreeBSD 4.7 and Python
    # 2.2.1, this method completed in 0.95s, and 0.05s if there was no
    # scheme to find. By comparison,
    #   urllib.splittype()[0] took 1.5s always;
    #   Ft.Lib.Uri.split_uri_ref()[0] took 2.5s always;
    #   urlparse.urlparse()[0] took 3.5s always.
    m = SCHEME_PATTERN.match(uriRef)
    if m is None:
        return None
    else:
        return m.group(1)


def strip_fragment(uriRef):
    """
    Returns the given URI or URI reference with the fragment component, if
    any, removed.
    """
    return split_fragment(uriRef)[0]


def is_absolute(identifier):
    """
    Given a string believed to be a URI or URI reference, tests that it is
    absolute (as per RFC 3986), not relative -- i.e., that it has a scheme.
    """
    # We do it this way to avoid compiling another massive regex.
    return get_scheme(identifier) is not None


_ntPathToUriSetupCompleted = False
def _initNtPathPattern():
    """
    Called internally to compile the regular expression used by
    os_path_to_uri() on Windows just once, the first time the function is
    called.
    """
    global _ntPathToUriSetupCompleted
    if _ntPathToUriSetupCompleted:
        return
    # path variations we try to handle:
    #
    # a\b\c (a relative path)
    #    file:a/b/c is the best we can do.
    #    Dot segments should not be collapsed in the final URL.
    #
    # \a\b\c
    #    file:///a/b/c is correct
    #
    # C:\a\b\c
    #    urllib.urlopen() requires file:///C|/a/b/c or ///C|/a/b/c
    #     because it currently relies on urllib.url2pathname().
    #    Windows resolver will accept the first or file:///C:/a/b/c
    #
    # \\host\share\x\y\z
    #    Windows resolver accepts file://host/share/x/y/z
    #    Netscape (4.x?) accepts file:////host/share/x/y/z
    #
    # If an entire drive is shared, the share name might be
    #  $drive$, like this: \\host\$c$\a\b\c
    #  We could recognize it as a drive letter, but it's probably
    #  best not to treat it specially, since it will never occur
    #  without a host. It's just another share name.
    #
    # There's also a weird C:\\host\share\x\y\z convention
    #  that is hard to find any information on. Presumably the C:
    #  is ignored, but the question is do we put it in the URI?
    #
    # So the format, in ABNF, is roughly:
    # [ drive ":" ] ( [ "\\" host "\" share ] abs-path ) / rel-path
    drive         = r'(?P<drive>[A-Za-z])'
    host          = r'(?P<host>[^\\]*)'
    share         = r'(?P<share>[^\\]+)'
    abs_path      = r'(?P<abspath>\\(?:[^\\]+\\?)*)'
    rel_path      = r'(?P<relpath>(?:[^\\]+\\?)*)'
    NT_PATH_REGEX = r"^(?:%s:)?(?:(?:(?:\\\\%s\\%s)?%s)|%s)$" % (
                        drive,
                        host,
                        share,
                        abs_path,
                        rel_path)
    global NT_PATH_PATTERN
    NT_PATH_PATTERN = re.compile(NT_PATH_REGEX)
    # We can now use NT_PATH_PATTERN.match(path) to parse the path and use
    #  the returned object's .groupdict() method to get a dictionary of
    #  path subcomponents. For example,
    #  NT_PATH_PATTERN.match(r"\\h\$c$\x\y\z").groupdict()
    #  yields
    #  {'abspath': r'\x\y\z',
    #   'share': '$c$',
    #   'drive': None,
    #   'host': 'h',
    #   'relpath': None
    #  }
    # Note that invalid paths such as r'\\foo\bar'
    #  (a UNC path with no trailing '\') will not match at all.
    _ntPathToUriSetupCompleted = True
    return


def _splitNtPath(path):
    """
    Called internally to get a tuple representing components of the given
    Windows path.
    """
    if not _ntPathToUriSetupCompleted:
        _initNtPathPattern()
    m = NT_PATH_PATTERN.match(path)
    if not m:
        raise ValueError("Path {path} is not a valid Windows path.".format(path=path))
    components = m.groupdict()
    (drive, host, share, abspath, relpath) = (
        components['drive'],
        components['host'],
        components['share'],
        components['abspath'],
        components['relpath'],
        )
    return (drive, host, share, abspath, relpath)


def _get_drive_letter(s):
    """
    Called internally to get a drive letter from a string, if the string
    is a drivespec.
    """
    if len(s) == 2 and s[1] in ':|' and s[0] in ascii_letters:
        return s[0]
    return


def os_path_to_uri(path, attemptAbsolute=True, osname=None):
    r"""This function converts an OS-specific file system path to a URI of
    the form 'file:///path/to/the/file'.

    In addition, if the path is absolute, any dot segments ('.' or '..') will
    be collapsed, so that the resulting URI can be safely used as a base URI
    by functions such as absolutize().

    The given path will be interpreted as being one that is appropriate for
    use on the local operating system, unless a different osname argument is
    given.

    If the given path is relative, an attempt may be made to first convert
    the path to absolute form by interpreting the path as being relative
    to the current working directory.  This is the case if the attemptAbsolute
    flag is True (the default).  If attemptAbsolute is False, a relative
    path will result in a URI of the form file:relative/path/to/a/file .

    attemptAbsolute has no effect if the given path is not for the
    local operating system.

    On Windows, the drivespec will become the first step in the path component
    of the URI. If the given path contains a UNC hostname, this name will be
    used for the authority component of the URI.

    Warning: Some libraries, such as urllib.urlopen(), may not behave as
    expected when given a URI generated by this function. On Windows you may
    want to call re.sub('(/[A-Za-z]):', r'\1|', uri) on the URI to prepare it
    for use by functions such as urllib.url2pathname() or urllib.urlopen().

    This function is similar to urllib.pathname2url(), but is more featureful
    and produces better URIs.
    """
    # Problems with urllib.pathname2url() on all platforms include:
    # - the generated URL has no scheme component;
    # - percent-encoding is incorrect, due to urllib.quote() issues.
    #
    # Problems with urllib.pathname2url() on Windows include:
    # - trailing backslashes are ignored;
    # - all leading backslashes are considered part of the absolute
    #    path, so UNC paths aren't properly converted (assuming that
    #    a proper conversion would be to use the UNC hostname in the
    #    hostname component of the resulting URL);
    # - non-leading, consecutive backslashes are collapsed, which may
    #    be desirable but is correcting what is, arguably, user error;
    # - the presence of a letter followed by ":" is believed to
    #    indicate a drivespec, no matter where it occurs in the path,
    #    which may have been considered a safe assumption since the
    #    drivespec is the only place where ":" can legally, but there's
    #    no need to search the whole string for it;
    # - the ":" in a drivespec is converted to "|", a convention that
    #    is becoming increasingly less common. For compatibility, most
    #    web browser resolvers will accept either "|" or ":" in a URL,
    #    but urllib.urlopen(), relying on url2pathname(), expects "|"
    #    only. In our opinion, the callers of those functions should
    #    ensure that the arguments are what are expected. Our goal
    #    here is to produce a quality URL, not a URL designed to play
    #    nice with urllib's bugs & limitations.
    # - it treats "/" the same as "\", which results in being able to
    #    call the function with a posix-style path, a convenience
    #    which allows the caller to get sloppy about whether they are
    #    really passing a path that is apprropriate for the desired OS.
    #    We do this a lot in 4Suite.
    #
    # There is some disagreement over whether a drivespec should be placed in
    # the authority or in the path. Placing it in the authority means that
    # ":", which has a reserved purpose in the authority, cannot be used --
    # this, along with the fact that prior to RFC 3986, percent-encoded
    # octets were disallowed in the authority, is presumably a reason why "|"
    # is a popular substitute for ":". Using the authority also allows for
    # the drive letter to be retained whe resolving references like this:
    #   reference '/a/b/c' + base 'file://C|/x/y/z' = 'file://C|/a/b/c'
    # The question is, is that really the ideal result? Should the drive info
    # be inherited from the base URI, if it is unspecified in a reference
    # that is otherwise representing an absolute path? Using the authority
    # for this purpose means that it would be overloaded if we also used it
    # to represent the host part of a UNC path. The alternative is to put the
    # UNC host in the path (e.g. 'file:////host/share/path'), but when such a
    # URI is used as a base URI, relative reference resolution often returns
    # unexpected results.
    #
    osname = osname or os.name

    if osname == 'nt':
        if WINDOWS_SLASH_COMPAT:
            path = path.replace('/','\\')
        (drive, host, share, abspath, relpath) = _splitNtPath(path)
        if attemptAbsolute and relpath is not None and osname == os.name:
            path = os.path.join(os.getcwd(), relpath)
            (drive, host, share, abspath, relpath) = _splitNtPath(path)
        path = abspath or relpath
        path = '/'.join([percent_encode(seg) for seg in path.split('\\')])
        uri = 'file:'
        if host:
            uri += '//%s' % percent_encode(host)
        elif abspath:
            uri += '//'
        if drive:
            uri += '/%s:' % drive.upper()
        if share:
            uri += '/%s' % percent_encode(share)
        if abspath:
            path = remove_dot_segments(path)
        uri += path

    elif osname == 'posix':
        try:
            from posixpath import isabs
        except ImportError:
            isabs = lambda p: p[:1] == '/'
        pathisabs = isabs(path)
        if pathisabs:
            path = remove_dot_segments(path)
        elif attemptAbsolute and osname == os.name:
            path = os.path.join(os.getcwd(), path)
            pathisabs = isabs(path)
        path = '/'.join([percent_encode(seg) for seg in path.split('/')])
        if pathisabs:
            uri = 'file://%s' % path
        else:
            uri = 'file:%s' % path

    else:
        # 4Suite only supports posix and nt, so we're not going to worry about
        # improving upon urllib.pathname2url() for other OSes.
        if osname == os.name:
            from urllib import pathname2url
            if attemptAbsolute and not os.path.isabs(path):
                path = os.path.join(os.getcwd(), path)
        else:
            try:
                module = '%surl2path' % osname
                exec('from %s import pathname2url' % module, globals(), locals())
            except ImportError:
                raise RuntimeError("Platform {platform} not supported by URI function "
                "{function}".format(platform=osname, function="os_path_to_uri"))
        uri = 'file:' + pathname2url(path)

    return uri


def uri_to_os_path(uri, attemptAbsolute=True, encoding='utf-8', osname=None):
    r"""
    This function converts a URI reference to an OS-specific file system path.

    If the URI reference is given as a Unicode string, then the encoding
    argument determines how percent-encoded components are interpreted, and
    the result will be a Unicode string. If the URI reference is a regular
    byte string, the encoding argument is ignored and the result will be a
    byte string in which percent-encoded octets have been converted to the
    bytes they represent. For example, the trailing path segment of
    'file:///a/b/%E2%80%A2' will by default be converted to '\u2022',
    because sequence E2 80 A2 represents character U+2022 in UTF-8. If the
    string were not Unicode, the trailing segment would become the 3-byte
    string '\xe2\x80\xa2'.

    The osname argument determines for what operating system the resulting
    path is appropriate. It defaults to os.name and is typically the value
    'posix' on Unix systems (including Mac OS X and Cygwin), and 'nt' on
    Windows NT/2000/XP.

    This function is similar to urllib.url2pathname(), but is more featureful
    and produces better paths.

    If the given URI reference is not relative, its scheme component must be
    'file', and an exception will be raised if it isn't.

    In accordance with RFC 3986, RFC 1738 and RFC 1630, an authority
    component that is the string 'localhost' will be treated the same as an
    empty authority.

    Dot segments ('.' or '..') in the path component are NOT collapsed.

    If the path component of the URI reference is relative and the
    attemptAbsolute flag is True (the default), then the resulting path
    will be made absolute by considering the path to be relative to the
    current working directory. There is no guarantee that such a result
    will be an accurate interpretation of the URI reference.

    attemptAbsolute has no effect if the
    result is not being produced for the local operating system.

    Fragment and query components of the URI reference are ignored.

    If osname is 'posix', the authority component must be empty or just
    'localhost'. An exception will be raised otherwise, because there is no
    standard way of interpreting other authorities. Also, if '%2F' is in a
    path segment, it will be converted to r'\/' (a backslash-escaped forward
    slash). The caller may need to take additional steps to prevent this from
    being interpreted as if it were a path segment separator.

    If osname is 'nt', a drivespec is recognized as the first occurrence of a
    single letter (A-Z, case-insensitive) followed by '|' or ':', occurring as
    either the first segment of the path component, or (incorrectly) as the
    entire authority component. A UNC hostname is recognized as a non-empty,
    non-'localhost' authority component that has not been recognized as a
    drivespec, or as the second path segment if the first path segment is
    empty. If a UNC hostname is detected, the result will begin with
    '\\<hostname>\'. If a drivespec was detected also, the first path segment
    will be '$<driveletter>$'. If a drivespec was detected but a UNC hostname
    was not, then the result will begin with '<driveletter>:'.

    Windows examples:
    'file:x/y/z' => r'x\y\z';
    'file:/x/y/z' (not recommended) => r'\x\y\z';
    'file:///x/y/z' => r'\x\y\z';
    'file:///c:/x/y/z' => r'C:\x\y\z';
    'file:///c|/x/y/z' => r'C:\x\y\z';
    'file:///c:/x:/y/z' => r'C:\x:\y\z' (bad path, valid interpretation);
    'file://c:/x/y/z' (not recommended) => r'C:\x\y\z';
    'file://host/share/x/y/z' => r'\\host\share\x\y\z';
    'file:////host/share/x/y/z' => r'\\host\share\x\y\z'
    'file://host/x:/y/z' => r'\\host\x:\y\z' (bad path, valid interp.);
    'file://localhost/x/y/z' => r'\x\y\z';
    'file://localhost/c:/x/y/z' => r'C:\x\y\z';
    'file:///C:%5Cx%5Cy%5Cz' (not recommended) => r'C:\x\y\z'
    """
    (scheme, authority, path) = split_uri_ref(uri)[0:3]
    if scheme and scheme != 'file':
        raise ValueError("Only a 'file' URI can be converted to an OS-specific path; "
                "URI given was {uri}".format(uri=uri))
    # enforce 'localhost' URI equivalence mandated by RFCs 1630, 1738, 3986
    if authority == 'localhost':
        authority = None
    osname = osname or os.name

    if osname == 'nt':
        # Get the drive letter and UNC hostname, if any. Fragile!
        unchost = None
        driveletter = None
        if authority:
            authority = percent_decode(authority, encoding=encoding)
            if _get_drive_letter(authority):
                driveletter = authority[0]
            else:
                unchost = authority
        if not (driveletter or unchost):
            # Note that we have to treat %5C (backslash) as a path separator
            # in order to catch cases like file:///C:%5Cx%5Cy%5Cz => C:\x\y\z
            # We will also treat %2F (slash) as a path separator for
            # compatibility.
            if WINDOWS_SLASH_COMPAT:
                regex = '%2[fF]|%5[cC]'
            else:
                regex = '%5[cC]'
            path = re.sub(regex, '/', path)
            segs = path.split('/')
            if not segs[0]:
                # //host/... => [ '', '', 'host', '...' ]
                if len(segs) > 2 and not segs[1]:
                    unchost = percent_decode(segs[2], encoding=encoding)
                    path = len(segs) > 3 and '/' + '/'.join(segs[3:]) or ''
                # /C:/...    => [ '', 'C:', '...' ]
                elif len(segs) > 1:
                    driveletter = _get_drive_letter(percent_decode(segs[1],
                                                   encoding=encoding))
                    if driveletter:
                        path = len(segs) > 2 and '/' + '/'.join(segs[2:]) or ''
            else:
                # C:/...     => [ 'C:', '...' ]
                driveletter = _get_drive_letter(percent_decode(segs[0],
                                                encoding=encoding))
                if driveletter:
                    path = len(segs) > 1 and path[2:] or ''



        # Do the conversion of the path part
        sep = '\\' # we could try to import from ntpath,
                   # but at this point it would just waste cycles.
        path = percent_decode(path.replace('/', sep), encoding=encoding)

        # Assemble and return the path
        if unchost:
            # It's a UNC path of the form \\host\share\path.
            # driveletter is ignored.
            path = r'%s%s%s' % (sep * 2, unchost, path)
        elif driveletter:
            # It's an ordinary Windows path of the form C:\x\y\z
            path = r'%s:%s' % (driveletter.upper(), path)
        # It's an ordinary Windows path of the form \x\y\z or x\y\z.
        # We need to make sure it doesn't end up looking like a UNC
        # path, so we discard extra leading backslashes
        elif path[:1] == '\\':
            path = re.sub(r'^\\+', '\\\\', path)
        # It's a relative path. If the caller wants it absolute, attempt to comply
        elif attemptAbsolute and osname == os.name:
            path = os.path.join(os.getcwd(), path)

        return path

    elif osname == 'posix':
        # a non-empty, non-'localhost' authority component is ambiguous on Unix
        if authority:
            raise ValueError("A URI containing a remote host name cannot be converted to a "
                " path on posix; URI given was {uri}".format(uri=uri))
        # %2F in a path segment would indicate a literal '/' in a
        # filename, which is possible on posix, but there is no
        # way to consistently represent it. We'll backslash-escape
        # the literal slash and leave it to the caller to ensure it
        # gets handled the way they want.
        path = percent_decode(re.sub('%2[fF]', '\\/', path), encoding=encoding)
        # If it's relative and the caller wants it absolute, attempt to comply
        if attemptAbsolute and osname == os.name and not os.path.isabs(path):
            path = os.path.join(os.getcwd(), path)

        return path

    else:
        # 4Suite only supports posix and nt, so we're not going to worry about
        # improving upon urllib.pathname2url() for other OSes.
        if osname == os.name:
            from urllib import url2pathname
        else:
            try:
                module = '%surl2path' % osname
                exec('from %s import url2pathname' % module, globals(), locals())
            except ImportError:
                raise RuntimeError("Platform {platform} not supported by URI function "
                "{function}".format(platform=osname, function="uri_to_os_path"))
        # drop the scheme before passing to url2pathname
        if scheme:
            uri = uri[len(scheme)+1:]
        return url2pathname(uri)

REG_NAME_HOST_PATTERN = re.compile(r"^(?:(?:[0-9A-Za-z\-_\.!~*'();&=+$,]|(?:%[0-9A-Fa-f]{2}))*)$")


def path_resolve(paths):
    """
    This function takes a list of file URIs.  The first can be
    absolute or relative to the URI equivalent of the current working
    directory. The rest must be relative to the first.
    The function converts them all to OS paths appropriate for the local
    system, and then creates a single final path by resolving each path
    in the list against the following one. This final path is returned
    as a URI.
    """
    if not paths: return paths
    paths = [uri_to_os_path(p, attemptAbsolute=False) for p in paths]
    if not os.path.isabs(paths[0]):
        paths[0] = os.path.join(os.getcwd(), paths[0])
    resolved = reduce(lambda a, b: \
                       basejoin(os.path.isdir(a)
                                 and os_path_to_uri(
                                      os.path.join(a, ''),
                                      attemptAbsolute=False,
                                     ) or os_path_to_uri(a, attemptAbsolute=False),
                                os_path_to_uri(b, attemptAbsolute=False)[5:]),
                       paths)
    return resolved


def basejoin(base, uriRef):
    """
    Merges a base URI reference with another URI reference, returning a
    new URI reference.

    It behaves exactly the same as absolutize(), except the arguments
    are reversed, and it accepts any URI reference (even a relative URI)
    as the base URI. If the base has no scheme component, it is
    evaluated as if it did, and then the scheme component of the result
    is removed from the result, unless the uriRef had a scheme. Thus, if
    neither argument has a scheme component, the result won't have one.

    This function is named basejoin because it is very much like
    urllib.basejoin(), but it follows the current RFC 3986 algorithms
    for path merging, dot segment elimination, and inheritance of query
    and fragment components.

    WARNING: This function exists for 2 reasons: (1) because of a need
    within the 4Suite repository to perform URI reference absolutization
    using base URIs that are stored (inappropriately) as absolute paths
    in the subjects of statements in the RDF model, and (2) because of
    a similar need to interpret relative repo paths in a 4Suite product
    setup.xml file as being relative to a path that can be set outside
    the document. When these needs go away, this function probably will,
    too, so it is not advisable to use it.
    """
    if is_absolute(base):
        return absolutize(uriRef, base)
    else:
        dummyscheme = 'basejoin'
        res = absolutize(uriRef, '%s:%s' % (dummyscheme, base))
        if is_absolute(uriRef):
            # scheme will be inherited from uriRef
            return res
        else:
            # no scheme in, no scheme out
            return res[len(dummyscheme)+1:]


def join(*uriparts):
    """
    Merges a series of URI reference parts, returning a new URI reference.

    Much like iri.basejoin, but takes multiple arguments
    """
    if len(uriparts) == 0:
        raise TypeError("FIXME...")
    elif len(uriparts) == 1:
        return uriparts[0]
    else:
        base = uriparts[0]
        for part in uriparts[1:]:
            base = basejoin(base.rstrip(DEFAULT_HIERARCHICAL_SEP) + DEFAULT_HIERARCHICAL_SEP, part)
        return base


#generate_iri
#Use:
#from uuid import *; newuri = uuid4().urn

#=======================================================================
#
# Further reading re: percent-encoding
#
# http://lists.w3.org/Archives/Public/ietf-http-wg/2004JulSep/0009.html
#
#=======================================================================
#
# 'file:' URI notes
#
# 'file:' URI resolution is difficult to get right, because the 'file'
# URL scheme is underspecified, and is handled by resolvers in very
# lenient and inconsistent ways.
#
# RFC 3986 provides definitive clarification on how all URIs,
# including the quirky 'file:' ones, are to be interpreted for purposes
# of resolution to absolute form, so that is what we implement to the
# best of our ability.
#
#-----------------------------------------------------------------------
#
# Notes from our previous research on 'file:' URI resolution:
#
# According to RFC 2396 (original), these are valid absolute URIs:
#  file:/autoexec.bat     (scheme ":" abs_path)
#  file:///autoexec.bat   (scheme ":" net_path)
#
# This one is valid but is not what you'd expect it to be:
#
#  file://autoexec.bat    (authority = autoexec.bat, no abs_path)
#
# If you have any more than 3 slashes, it's OK because each path segment
# can be an empty string.
#
# This one is valid too, although everything after 'file:' is
# considered an opaque_part (note that RFC 3986 changes this):
#
#   file:etc/passwd
#
# Unescaped backslashes are NOT allowed in URIs, ever.
# It is not possible to use them as path segment separators.
# Yet... Windows Explorer will accept these:
#   file:C:\WINNT\setuplog.txt
#   file:/C:\WINNT\setuplog.txt
#   file:///C:\WINNT\setuplog.txt
# However, it will also accept "|" in place of the colon after the drive:
#   file:C|/WINNT/setuplog.txt
#   file:/C|/WINNT/setuplog.txt
#   file:///C|/WINNT/setuplog.txt
#
# RFC 1738 says file://localhost/ and file:/// are equivalent;
# localhost in this case is always the local machine, no matter what
# your DNS says.
#
# Basically, any file: URI is valid. Good luck resolving them, though.
#
# Jeremy's idea is to not use open() or urllib.urlopen() on Windows;
# instead, use a C function that wraps Windows' generic open function,
# which resolves any path or URI exactly as Explorer would (he thinks).
#
#-----------------------------------------------------------------------
#
# References for further research on 'file:' URI resolution:
#  http://mail.python.org/pipermail/xml-sig/2001-February/004572.html
#  http://mail.python.org/pipermail/xml-sig/2001-March/004791.html
#  http://mail.python.org/pipermail/xml-sig/2002-August/008236.html
#  http://www.perldoc.com/perl5.8.0/lib/URI/file.html
#  http://lists.w3.org/Archives/Public/uri/2004Jul/0013.html
#
#=======================================================================



import sys
import ssl
import asyncio


try:
    import aiohttp
    from aiohttp import ClientSession

    AIOHTTP_ERROR_MENAGERIE = (
        ssl.CertificateError,
        asyncio.TimeoutError,
        aiohttp.ClientOSError,
        aiohttp.ServerDisconnectedError,
        aiohttp.ClientResponseError,
        aiohttp.ClientError,
    )
except ImportError:
    pass


def go_async(launch_task, close_loop=False):
    '''
    Convenience function to launch a coroutine asynchronously
    Basically Python 3.7's asyncio.run() simulated for 3.5 & 3.6
    
    WARNING: You can generally only use this once per interpreter session, unless you set close_loop=False

    >>> import asyncio
    >>> from amara3.asynctools import go_async
    >>> async def x():
    ...     await asyncio.sleep(1)
    ...     return 'ndewo'
    >>> retval = go_async(x())
    >>> retval
    'ndewo'
    '''
    loop = asyncio.get_event_loop()
    resp = loop.run_until_complete(launch_task)
    if close_loop: loop.close()
    return resp

#async def progress_indicator docstring rendered here for easy testing, for now
'''
import sys
import asyncio
from amara3.asynctools import progress_indicator, go_async
async def x():
    print('1', end='')
    await asyncio.sleep(2)
    print('2', end='')
    await asyncio.sleep(2)
    print('3', end='')

_ = go_async(asyncio.gather(x(), progress_indicator(0.5)))
'''

async def progress_indicator(delay, loop=None, out=sys.stdout, max_width=80):
    '''
    Coroutine useful for progress indication to console,
    printing dots when scheduled, after a given delay

    >>> import sys
    >>> import asyncio
    >>> from amara3.asynctools import progress_indicator, go_async
    >>> async def x():
    ...     print('1', end='')
    ...     await asyncio.sleep(2)
    ...     print('2', end='')
    ...     await asyncio.sleep(2)
    ...     print('3', end='')
    >>> _ = go_async(asyncio.gather(x(), progress_indicator(0.5)))
    1...2....3.>>> 
    '''
    if not loop: loop = asyncio.get_event_loop()
    count_to_width = 0
    while True:
        try:
            await asyncio.sleep(delay)
        except asyncio.CancelledError:
            break
        count_to_width += 1
        #Set a max width for th eline of dots & newline if reached
        if count_to_width > max_width:
            print('', file=sys.stderr)
            count_to_width = 0
        #Print a dot, with no newline afterward & force the output to appear immediately
        print('.', end='', file=out, flush=True)
        #Check if this is the last remaining task, and exit if so
        num_active_tasks = [ task for task in asyncio.Task.all_tasks(loop)
                                  if not task.done() ]
        if len(num_active_tasks) == 1:
            break


#class req_tracer docstring rendered here for easy testing, for now
'''
import aiohttp
from amara3.asynctools import go_async, req_tracer
rtimings = req_tracer()
async def access_site():
    url = 'http://artscene.textfiles.com/information/ascii-newmedia.txt'
    async with aiohttp.ClientSession(trace_configs=[rtimings.trace_config]) as sess:
        #If trace_request_ctx omitted, results stored in rtimings.request[None]
        async with sess.get(url, trace_request_ctx={'reqid': 'ID'}) as response:
            t = rtimings.request['ID'].get('total_elapsed', 0)
            t = '{:.3f}'.format(t) if t else 'UNKNOWN'
            print('Web request took', t, 'seconds')
            return response

resp = go_async(access_site())
'''

class req_tracer:
    '''
    aiohttp request tracer helper
    Requires aiohttp version 3.0.
    
    See: https://docs.aiohttp.org/en/stable/tracing_reference.html#aiohttp-client-tracing-reference
    
    >>> import aiohttp
    >>> from amara3.asynctools import go_async, req_tracer
    >>> rtimings = req_tracer()
    >>> async def access_site():
    ...     url = 'http://artscene.textfiles.com/information/ascii-newmedia.txt'
    ...     async with aiohttp.ClientSession(trace_configs=[rtimings.trace_config]) as sess:
    ...         #If trace_request_ctx omitted, results stored in rtimings.request[None]
    ...         async with sess.get(url, trace_request_ctx={'reqid': 'ID'}) as response:
    ...             t = rtimings.request['ID'].get('total_elapsed', 0)
    ...             t = '{:.3f}'.format(t) if t else 'UNKNOWN'
    ...             print('Web request took', t, 'seconds')
    ...             return response
    ... 
    >>> resp = go_async(access_site())
    Web request took 0.121 seconds
    >>> 
    '''
    def __init__(self):
        #Store the tracked timings according to a key specified in the request context (or use None key if none provided)
        self.request = {}
        self.trace_config = aiohttp.TraceConfig()
 
        self.trace_config.on_request_start.append(self.start_t)
        self.trace_config.on_request_redirect.append(self.redirected)
        self.trace_config.on_dns_resolvehost_start.append(self.dns_start_t)
        self.trace_config.on_dns_resolvehost_end.append(self.dns_end_t)
        self.trace_config.on_connection_create_start.append(self.connect_start_t)
        self.trace_config.on_connection_create_end.append(self.connect_end_t)
        self.trace_config.on_request_end.append(self.end_t)
        self.trace_config.on_request_chunk_sent.append(self.chunk_sent)
        self.trace_config.on_response_chunk_received.append(self.chunk_received)

    async def start_t(self, session, context, params):
        time = session.loop.time()
        context.start_t = time
        context.is_redirect = False
 
    async def connect_start_t(self, session, context, params):
        time = session.loop.time()
        since_start = time - context.start_t
        context.connect_start_t = since_start
 
    async def redirected(self, session, context, params):
        time = session.loop.time()
        since_start = time - context.start_t
        context.redirected = since_start
        context.is_redirect = True
 
    async def dns_start_t(self, session, context, params):
        time = session.loop.time()
        since_start = time - context.start_t
        context.dns_start_t = since_start
 
    async def dns_end_t(self, session, context, params):
        time = session.loop.time()
        since_start = time - context.start_t
        context.dns_end_t = since_start
 
    async def connect_end_t(self, session, context, params):
        time = session.loop.time()
        since_start = time - context.start_t
        context.connect_end_t = since_start
     
    async def chunk_sent(self, session, context, params):
        time = session.loop.time()
        since_start = time - context.start_t
        context.chunk_sent = since_start
     
    async def chunk_received(self, session, context, params):
        time = session.loop.time()
        since_start = time - context.start_t
        context.chunk_received = since_start
 
    async def end_t(self, session, context, params):
        if not context.trace_request_ctx: return
        #Store the tracked timings according to a key specified in the request context (or use None key if none provided)
        reqid = context.trace_request_ctx.get('reqid')
        time_now = session.loop.time()
        total_elapsed = time_now - context.start_t
        context.end_t = total_elapsed
 
        #The DNS & connect callbacks are not getting called, for some reason
        #dns_lookup_and_dial_time = context.dns_end_t - context.dns_start_t
        #connect_time = context.connect_end_t - dns_lookup_and_dial_time
        #transfer_time = total_elapsed - context.connect_end_t
        is_redirect = context.is_redirect
        
        self.request.setdefault(reqid, {}).update({
            #'dns_lookup_and_dial_time': dns_lookup_and_dial_time,
            #'connect_time': connect_time,
            #'transfer_time': transfer_time,
            'is_redirect': is_redirect,
            'total_elapsed': total_elapsed
        })




import pytest
import io, os, sys, inspect #,codecs
import warnings
from amara3 import inputsource, iri
from amara3.iri import IriError
from amara3.inputsource import factory

import os, inspect
def module_path(local_function):
   ''' returns the module path without the use of __file__.  Requires a function defined 
   locally in the module.
   from http://stackoverflow.com/questions/729583/getting-file-path-of-imported-module'''
   return os.path.abspath(inspect.getsourcefile(local_function))

#hack to locate test resource (data) files regardless of from where nose was run
RESOURCEPATH = os.path.normpath(os.path.join(module_path(lambda _: None), '../resource/'))


def test_string_is():
    inp = inputsource('abc')
    assert inp.stream.__class__ == io.StringIO
    assert inp.iri is None
    assert inp.stream.read() == 'abc'


def test_factory_string_is():
    inpl = factory('abc')
    for inp in inpl:
        assert inp.stream.__class__ == io.StringIO
        assert inp.iri is None
        assert inp.stream.read() == 'abc'


def test_stringlist_is():
    inpl = factory(['abc', 'def', 'ghi'])
    inp = inpl[0]
    assert inp.stream.__class__ == io.StringIO
    assert inp.iri is None
    assert inp.stream.read() == 'abc'
    inp = inpl[1]
    assert inp.stream.__class__ == io.StringIO
    assert inp.iri is None
    assert inp.stream.read() == 'def'
    inp = inpl[2]
    assert inp.stream.__class__ == io.StringIO
    assert inp.iri is None
    assert inp.stream.read() == 'ghi'


def test_file_is():
    fname = os.path.join(RESOURCEPATH, 'spam.txt')
    inp = inputsource(open(fname))
    assert inp.iri is None
    assert inp.stream.read() == 'monty\n'


def test_factory_file_is():
    fname = os.path.join(RESOURCEPATH, 'spam.txt')
    inpl = factory(open(fname))
    for inp in inpl:
        assert inp.iri is None
        assert inp.stream.read() == 'monty\n'


def test_stringio_is():
    inp = inputsource(io.StringIO('abc'))
    assert inp.iri is None
    assert inp.stream.read() == 'abc'


def test_bytesio_is():
    inp = inputsource(io.BytesIO('abc'.encode('utf-8')))
    assert inp.iri is None
    assert inp.stream.read() == b'abc'


def test_filelist_is():
    files = [ open(os.path.join(RESOURCEPATH, f)) for f in ('spam.txt', 'eggs.txt') ]
    inpl = factory(files)
    inp = inpl[0]
    assert inp.iri is None
    assert inp.stream.read() == 'monty\n'
    inp = inpl[1]
    assert inp.iri is None
    assert inp.stream.read() == 'python\n'


def test_zip_is():
    zf = open(os.path.join(RESOURCEPATH, 'speggs.zip'), 'rb')
    inpl = factory(zf, zipcheck=True)
    inp = next(inpl)
    assert inp.iri is None
    assert inp.stream.read() == b'python\n'
    inp = next(inpl)
    assert inp.iri is None
    assert inp.stream.read() == b'monty\n'




# amara3.inputsource
"""
Utilities for managing data inputs for Amara

Copyright 2008-2015 Uche Ogbuji
"""

import zipfile
import functools
from enum import Enum
from io import StringIO, BytesIO

from amara3 import iri
from urllib.request import urlopen

class inputsourcetype(Enum):
    unknown = 0
    stream = 1
    string = 2
    iri = 3
    filename = 4
    zipfilestream = 5


def factory(obj, defaultsourcetype=inputsourcetype.unknown, encoding=None, streamopenmode='rb', zipcheck=False):
    '''
    Helper function to create an iterable of inputsources from compound sources such as a zip file
    Returns an iterable of input sources

    obj - object, possibly list or tuple of items to be converted into one or more inputsource
    '''
    if isinstance(obj, inputsource):
        return obj
    _inputsource = functools.partial(inputsource, encoding=encoding, streamopenmode=streamopenmode)
    if isinstance(obj, tuple) or isinstance(obj, list):
        inputsources = [ _inputsource(o, sourcetype=defaultsourcetype) for o in obj ]
    #if isinstance(objs, str) or isinstance(objs, bytes) or isinstance(objs, bytearray):
    #Don't do a zipcheck unless we know we can rewind the obj
    #Because zipfile.is_zipfile fast forwards to EOF
    elif zipcheck and hasattr(obj, 'seek'):
        inputsources = []
        if zipfile.is_zipfile(obj):
            def zipfilegen():
                zf = zipfile.ZipFile(obj, 'r') #Mode must be r, w or a
                for info in zf.infolist():
                    #From the doc: Note If the ZipFile was created by passing in a file-like object as the first argument to the constructor, then the object returned by open() shares the ZipFile’s file pointer. Under these circumstances, the object returned by open() should not be used after any additional operations are performed on the ZipFile object.
                    yield(_inputsource(zf.open(info, mode='r')))
            inputsources = zipfilegen()
        else:
            #Because zipfile.is_zipfile fast forwards to EOF
            obj.seek(0, 0)
    else:
        inputsources = [_inputsource(obj)]
    return inputsources


class inputsource(object):
    '''
    A flexible class for managing input sources for e.g. XML processing

    Loosely based on Amara's old inputsource <https://github.com/zepheira/amara/blob/master/lib/lib/_inputsource.py>
    '''
    def __init__(self, obj, siri=None, encoding=None, streamopenmode='rb',
                    sourcetype=inputsourcetype.unknown):
        '''
        obj - byte string, proper string (only if you really know what you're doing),
            file-like object (stream), file path or URI.
        uri - optional override URI.  Base URI for the input source will be set to
            this value

        >>> from amara3 import inputsource
        >>> inp = inputsource('abc')
        >>> inp.stream
        <_io.StringIO object at 0x1056fbf78>
        >>> inp.iri
        >>> print(inp.iri)
        None
        >>> inp = inputsource(['abc', 'def']) #Now multiple streams in one source
        >>> inp.stream
        <_io.StringIO object at 0x1011aff78>
        >>> print(inp.iri)
        None
        >>> inp = next(inp)
        >>> inp.stream
        <_io.StringIO object at 0x1011af5e8>
        >>> print(inp.iri)
        None
        >>>
        '''
        # from amara3 import inputsource; inp = inputsource('foo.zip')
        # from amara3 import inputsource; inp = inputsource('test/resource/std-examples.zip')
        # s = inp.stream.read(100)
        # s
        # b'<?xml version="1.0" encoding="UTF-8"?>\r\n<!-- edited with XML Spy v4.3 U (http://www.xmlspy.com) by M'
        # s
        # b'<?xml version="1.0" encoding="UTF-8"?>\r\n<collection xmlns="http://www.loc.gov/MARC21/slim">\r\n  <reco'

        self.stream = None
        self.iri = siri
        self.sourcetype = sourcetype

        if obj in ('', b''):
            raise ValueError("Cannot parse an empty string as XML")

        if hasattr(obj, 'read'):
            #Create dummy Uri to use as base
            #uri = uri or uuid4().urn
            self.stream = obj
        #elif sourcetype == inputsourcetype.xmlstring:
            #See this article about XML detection heuristics
            #http://www.xml.com/pub/a/2007/02/28/what-does-xml-smell-like.html
            #uri = uri or uuid4().urn
        elif self.sourcetype == inputsourcetype.iri or (siri and iri.matches_uri_syntax(obj)):
            self.iri = siri or obj
            self.stream = urlopen(iri)
        elif self.sourcetype == inputsourcetype.filename or (siri and iri.is_absolute(obj) and not os.path.isfile(obj)):
            #FIXME: convert path to URI
            self.iri = siri or iri.os_path_to_uri(obj)
            self.stream = open(obj, streamopenmode)
        elif self.sourcetype == inputsourcetype.string or isinstance(obj, str) or isinstance(obj, bytes):
            self.stream = StringIO(obj)
            #If obj is beyond a certain length, don't even try it as a URI
            #if len(obj) < MAX_URI_LENGTH_FOR_HEURISTIC:
            #    self.iri = iri.os_path_to_uri(obj)
            #    self.stream = urlopen(siri)
        else:
            raise ValueError("Unable to recognize as an inputsource")
        return

    @staticmethod
    def text(obj, siri=None, encoding=None):
        '''
        Set up an input source from text, according to the markup convention of the term
        (i.e. in Python terms a string with XML, HTML, fragments thereof, or tag soup)

        Helps with processing content sources that are not unambiguously XML or HTML strings
        (e.g. could be mistaken for filenames or IRIs)
        '''
        return inputsource(obj, siri, encoding, sourcetype=inputsourcetype.string)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Highly recommend installing using `pip install .` not `python setup.py install`

Uses pkgutil-style namespace package (Working on figuring out PEP 420)

Note: careful not to conflate install_requires with requirements.txt

https://packaging.python.org/discussions/install-requires-vs-requirements/

Reluctantly use setuptools for now to get install_requires & long_description_content_type

$ python -c "import amara3; import amara3.iri; import amara3.uxml; import amara3.uxml.version; print(amara3.uxml.version.version_info)"
('3', '0', '1')
'''

import sys
from setuptools import setup
#from distutils.core import setup

PROJECT_NAME = 'amara3.iri'
PROJECT_DESCRIPTION = 'Module for handling Internationalized Resource Identifiers (IRIs). Core of the Amara3 project, which offers a variety of data processing tools.',
PROJECT_LICENSE = 'License :: OSI Approved :: Apache Software License'
PROJECT_AUTHOR = 'Uche Ogbuji'
PROJECT_AUTHOR_EMAIL = 'uche@ogbuji.net'
PROJECT_URL = 'https://github.com/uogbuji/amara3-xml'
PACKAGE_DIR = {'amara3': 'pylib'}
PACKAGES = [
    'amara3',
    'amara3.contrib',
]
SCRIPTS = []

#FIXME: Trim some of these as amara3-xml & versa setup.py files are updated to handle requirements
#See: requirements.txt
CORE_REQUIREMENTS = [
    'pytest',
]

# From http://pypi.python.org/pypi?%3Aaction=list_classifiers
CLASSIFIERS = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        #"Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet :: WWW/HTTP",
]

KEYWORDS=['web', 'data', 'url', 'uri', 'iri']

version_file = 'pylib/version.py'
exec(compile(open(version_file, "rb").read(), version_file, 'exec'), globals(), locals())
__version__ = '.'.join(version_info)

LONGDESC = '''Core of the Amara3 project, which provides a variety of data processing tools. Core is mostly a library for handling Internationalized Resource Identifiers (IRIs)
'''

LONGDESC_CTYPE = 'text/markdown',


setup(
    #namespace_packages=['amara3'],
    name=PROJECT_NAME,
    version=__version__,
    description=PROJECT_DESCRIPTION,
    license=PROJECT_LICENSE,
    author=PROJECT_AUTHOR,
    author_email=PROJECT_AUTHOR_EMAIL,
    #maintainer=PROJECT_MAINTAINER,
    #maintainer_email=PROJECT_MAINTAINER_EMAIL,
    url=PROJECT_URL,
    package_dir=PACKAGE_DIR,
    packages=PACKAGES,
    scripts=SCRIPTS,
    install_requires=CORE_REQUIREMENTS,
    classifiers=CLASSIFIERS,
    long_description=LONGDESC,
    long_description_content_type=LONGDESC_CTYPE,
    keywords=KEYWORDS,
)




"""
mem_check.py

functions for getting memoroy use of the current python process

Downloaded from: http://pythonchb.github.io/PythonTopics/weak_references.html

Windows and *nix versions

USAGE:

amount = get_mem_use(units='MB') # options are KB, MB, GB

"""

import sys

div = {'GB': 1024*1024*1024,
       'MB': 1024*1024,
       'KB': 1024,
       }

if sys.platform.startswith('win'):

    """

    Functions for getting memory usage of Windows processes.

    from:

    http://code.activestate.com/recipes/578513-get-memory-usage-of-windows-processes-using-getpro/

    get_mem_use(units='MB') is the one to get memory use for the current process.


    """
    import ctypes
    from ctypes import wintypes

    GetCurrentProcess = ctypes.windll.kernel32.GetCurrentProcess
    GetCurrentProcess.argtypes = []
    GetCurrentProcess.restype = wintypes.HANDLE

    SIZE_T = ctypes.c_size_t

    class PROCESS_MEMORY_COUNTERS_EX(ctypes.Structure):
        _fields_ = [
            ('cb', wintypes.DWORD),
            ('PageFaultCount', wintypes.DWORD),
            ('PeakWorkingSetSize', SIZE_T),
            ('WorkingSetSize', SIZE_T),
            ('QuotaPeakPagedPoolUsage', SIZE_T),
            ('QuotaPagedPoolUsage', SIZE_T),
            ('QuotaPeakNonPagedPoolUsage', SIZE_T),
            ('QuotaNonPagedPoolUsage', SIZE_T),
            ('PagefileUsage', SIZE_T),
            ('PeakPagefileUsage', SIZE_T),
            ('PrivateUsage', SIZE_T),
        ]

    GetProcessMemoryInfo = ctypes.windll.psapi.GetProcessMemoryInfo
    GetProcessMemoryInfo.argtypes = [
        wintypes.HANDLE,
        ctypes.POINTER(PROCESS_MEMORY_COUNTERS_EX),
        wintypes.DWORD,
    ]
    GetProcessMemoryInfo.restype = wintypes.BOOL

    def get_current_process():
        """Return handle to current process."""
        return GetCurrentProcess()

    def get_memory_info(process=None):
        """Return Win32 process memory counters structure as a dict."""
        if process is None:
            process = get_current_process()
        counters = PROCESS_MEMORY_COUNTERS_EX()
        ret = GetProcessMemoryInfo(process, ctypes.byref(counters),
                                   ctypes.sizeof(counters))
        if not ret:
            raise ctypes.WinError()
        info = dict((name, getattr(counters, name))
                    for name, _ in counters._fields_)
        return info

    def get_mem_use(units='MB'):
        """
        returns the total memory use of the current python process

        :param units='MB': the units you want the reslut in. Options are:
                           'GB', 'MB', 'KB'
        """
        info = get_memory_info()
        return info['PrivateUsage'] / float(div[units])


else: # for posix systems only tested on OS-X for now
    def get_mem_use(units='MB'):
        """
        returns the total memory use of the current python process

        :param units='MB': the units you want the reslut in. Options are:
                           'GB', 'MB', 'KB'
        """
        import resource
        #useage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        div = {'GB': 1024*1024*1024,
               'MB': 1024*1024,
               'KB': 1024,
               }
        return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / float(div[units])
